
""" Sample OBJ cube:
o tile-0-0-0

v 0.0 0.0 0.0
v 0.0 0.0 1.0
v 0.0 1.0 0.0
v 0.0 1.0 1.0
v 1.0 0.0 0.0
v 1.0 0.0 1.0
v 1.0 1.0 0.0
v 1.0 1.0 1.0

vn 0.0 0.0 1.0
vn 0.0 0.0 -1.0
vn 0.0 1.0 0.0
vn 0.0 -1.0 0.0
vn 1.0 0.0 0.0
vn -1.0 0.0 0.0
usemtl BLUE_MTL
s 1
f 1//2 7//2 5//2
f 1//2 3//2 7//2
f 1//6 4//6 3//6
f 1//6 2//6 4//6
f 3//3 8//3 7//3
f 3//3 4//3 8//3
f 5//5 7//5 8//5
f 5//5 8//5 6//5
f 1//4 5//4 6//4
f 1//4 6//4 2//4
f 2//1 6//1 8//1
f 2//1 8//1 4//1
"""

faces_list = [(1,7,5,2), (1,3,7,2),
              (1,4,3,6), (1,2,4,6),
              (3,8,7,3), (3,4,8,3),
              (5,7,8,5), (5,8,6,5),
              (1,5,6,4), (1,6,2,4),
              (2,6,8,1), (2,8,4,1)]


class ObjCube:
    vert_num = 0
    norm_num = 0
    def __init__(self, coords, color_name):
        self.start_vert = ObjCube.vert_num
        ObjCube.vert_num += 8
        self.start_norm = ObjCube.norm_num
        ObjCube.norm_num += 6
        self.coords = coords
        self.verts = self.get_verts(coords)
        self.normals = self.get_normals(coords)
        self.faces = self.get_faces()
        self.color_name = color_name

    def __str__(self):
        ret = f'o tile-{self.coords[0]}-{self.coords[1]}-{self.coords[2]}\n\n'
        for (x,y,z) in self.verts:
            ret += f'v {x}.0 {y}.0 {z}.0\n'
        ret += '\n'
        for (x,y,z) in self.normals:
            ret += f'vn {x} {y} {z}\n'
        ret += f'usemtl {self.color_name}_MTL\n'
        ret += 's 1\n'
        for (v1,v2,v3,n) in self.faces:
            ret += f'f {v1}//{n} {v2}//{n} {v3}//{n}\n'
        ret += '\n'
        return ret

    def __repr__(self):
        self.__str__()

    def get_verts(self, coords):
        verts = []
        for i in range(8):
            bits = f"{i:03b}"
            x = coords[0] + int(bits[0])
            y = coords[1] + int(bits[1])
            z = coords[2] + int(bits[2])
            verts.append((x,y,z))
        return verts

    def get_normals(self, coords):
        normals = [(0.0, 0.0, 1.0), (0.0, 0.0, -1.0),
                   (0.0, 1.0, 0.0), (0.0, -1.0, 0.0),
                   (1.0, 0.0, 0.0), (-1.0, 0.0, 0.0)]
        return normals

    def get_faces(self):
        faces = []
        for face in faces_list:
            faces.append((self.start_vert+face[0],
                          self.start_vert+face[1],
                          self.start_vert+face[2],
                          self.start_norm+face[3]))
        return faces


""" Sample OBJ material:
newmtl Blue_OBJ
Ns 225.000000
Ka 1.000000 1.000000 1.000000
Kd 0.000000 0.000000 1.000000  <- color, RBG
Ks 0.500000 0.500000 0.500000
Ke 0.000000 0.000000 0.000000
Ni 1.450000
d 1.000000
illum 2
"""

class ObjMaterial:
    def __init__(self, color_name, color_vector):
        self.color_name = color_name
        self.color_vector = color_vector

    def get_color_rgb(self):
        r = round(self.color_vector[0] / 255.0, 6)
        g = round(self.color_vector[1] / 255.0, 6)
        b = round(self.color_vector[2] / 255.0, 6)
        return f'{r} {g} {b}'

    def __str__(self):
        ret = f'newmtl {self.color_name}_MTL\n'
        ret += 'Ns 225.000000\n'
        ret += 'Ka 1.000000 1.000000 1.000000\n'
        ret += f'Kd {self.get_color_rgb()}\n'
        ret += 'Ks 0.500000 0.500000 0.500000\n'
        ret += 'Ke 0.000000 0.000000 0.000000\n'
        ret += 'Ni 1.450000\n'
        ret += 'd 1.000000\n'
        ret += 'illum 2\n'
        return ret

    def __repr__(self):
        return self.__str__()
