import wx
from SimulatorFramework import *
from SimulatorFramework.TileType import *
from PIL.ImageDraw2 import Pen
import svgwrite
from svgwrite import cm, mm, px
from svgwrite import pt as svgPt
from reportlab.pdfgen import canvas


N = North = 0
E = East = 1
S = South = 2
W = West = 3
U = Up = 4
D = Down = 5

Dirs = [N,E,S,W]
DirAbbrevArray = ("N", "E", "S", "W", "U", "D")
DirNameArray = ("NORTH", "EAST", "SOUTH", "WEST", "UP", "DOWN")

def GetFontSizeToFitTextInBounds(text, font, dc, width, height):
    retFont = font
    (w, pixelSize) = retFont.GetPixelSize()
    bContinue = True
    while bContinue == True:
        dc.SetFont(retFont)
        (w, h) = dc.GetTextExtent(text)
        if ((w <= width) and (h <= height)) or pixelSize <= 2:
            bContinue = False
        else:
            pixelSize = pixelSize - 2
            retFont.SetPixelSize((0, pixelSize))
    #(w, pixelSize) = retFont.GetPixelSize()
    #print("pixelSize: " + str(pixelSize))
    return retFont

###################################################

CELLSIZE = 180

# for 3D simulation
#----------------------------------------------------------
# in OpenGL 1.0 texture sides should be 2^n
TEXTURE_RES = 1024

# size of a side of a tile on the texture
SIDE_RES = int(TEXTURE_RES / 3)

# texture offsets of given side [N,E,S,W,U,D]
SIDE_OFFSET = [(0,0),(0,SIDE_RES),(SIDE_RES,0),(SIDE_RES,SIDE_RES),(2*SIDE_RES,0),(2*SIDE_RES,SIDE_RES)]
#----------------------------------------------------------

class PyTASTileType(TileType):
    """A representation of a tile type for use in the ISU TAS simulator.
       Every tile must have a unique name and a unique set of 4 glues. If
       a name is not explicitly provided, one will be formed automatically
       as a concatenation of the glues values of each side."""
    
    def __init__(self, name="", width=CELLSIZE, height=CELLSIZE, is3D=False, label=""):
        TileType.__init__(self, name, width, height, label)
        if is3D == False:
            self.dims = 2
            self.sides = {N:Glue(),E:Glue(),S:Glue(),W:Glue()}
        else:
            self.dims = 3
            self.sides = {N:Glue(),E:Glue(),S:Glue(),W:Glue(),U:Glue(),D:Glue()}
           
    def SetDimensions(self, dims):
        if self.dims == dims:
            return
        elif dims == 3:
            self.dims = 3
            self.sides[U] = Glue()
            self.sides[D] = Glue()
        elif dims == 2:
            self.dims = 2
            savedSides = self.sides
            self.sides = {N:savedSides[N], E:savedSides[E], S:savedSides[S], W:savedSides[W]}
        else:
            print("ERROR: invalid number of tile type dimensions: " + str(dims))

    def MakeImage(self, backgroundColor=None, mask=DISPLAY_ALL, pdfCanvas=None, svgCanvas=None, trans=None):
            w = int(CELLSIZE)
            h = int(CELLSIZE)
            bmp = wx.Bitmap(w, h)
            dc = wx.MemoryDC()
            dc.SelectObject(bmp)
            
            if not backgroundColor is None:
                bg = wx.Brush(backgroundColor)#(self.GetBackgroundColour())
                dc.SetBackground(bg)
            dc.Clear()

            brush = wx.Brush(self.colorVector)
            dc.SetBrush(brush)
            if not pdfCanvas is None:
                pdfColor = (float(self.colorVector[0])/255, float(self.colorVector[1])/255, float(self.colorVector[2])/255)
                pdfCanvas.setFillColorRGB(pdfColor[0], pdfColor[1], pdfColor[2])
            elif not svgCanvas is None:
                units = px
                svgColor = "rgb(" + str(self.colorVector[0]) + "," + str(self.colorVector[1]) + "," + str(self.colorVector[2]) + ")"
                #TODO: fix: svgGroup = svgCanvas.add(svgCanvas.g(id=self.GetName()))
                svgGroup = svgCanvas.add(svgCanvas.g())
            
            black = wx.TheColourDatabase.Find("Black")
            pen = wx.Pen(black, width=3)
            dc.SetPen(pen)

            tabSize = 10
            dc.DrawRectangle(0+tabSize, 0+tabSize, w-(2*tabSize), h-(2*tabSize))
            if not pdfCanvas is None:
                pdfCanvas.rect(0+tabSize, 0+tabSize, w-(2*tabSize), h-(2*tabSize), stroke=1, fill=1)
            elif not svgCanvas is None:
                svgGroup.add(svgCanvas.rect(insert=(trans[0]+tabSize, trans[1]+tabSize),
                                         size=(w-(2*tabSize), h-(2*tabSize)),
                                         fill=svgColor, stroke='black', stroke_width=5))
            brush = wx.Brush(black)
            dc.SetBrush(brush)
            if not pdfCanvas is None:
                pdfCanvas.setFillColor("black")

            # Draw northern tabs
            numTabs = self.sides[N].strength
            if numTabs > 0:
                spacing = w / (numTabs+1)
                currLoc = spacing
                for n in range(numTabs):
                    dc.DrawRectangle(int(currLoc-(tabSize/2)), 0, tabSize, tabSize)
                    if not pdfCanvas is None:
                        pdfCanvas.rect(currLoc-(tabSize/2), h-tabSize, tabSize, tabSize, stroke=1, fill=1)
                    elif not svgCanvas is None:
                        svgGroup.add(svgCanvas.rect(insert=(trans[0]+currLoc-(tabSize/2), trans[1]),
                         size=(tabSize, tabSize), fill='black', stroke='black', stroke_width=0))
                    currLoc = currLoc + spacing

            # Draw southern tabs
            numTabs = self.sides[S].strength
            if numTabs > 0:
                spacing = w / (numTabs+1)
                currLoc = spacing
                for n in range(numTabs):
                    dc.DrawRectangle(int(currLoc-(tabSize/2)), h-tabSize, tabSize, tabSize)
                    if not pdfCanvas is None:
                        pdfCanvas.rect(currLoc-(tabSize/2), 0, tabSize, tabSize, stroke=1, fill=1)
                    elif not svgCanvas is None:
                        svgGroup.add(svgCanvas.rect(insert=(trans[0]+currLoc-(tabSize/2), trans[1]+h-tabSize),
                         size=(tabSize, tabSize), fill='black', stroke='black', stroke_width=0))
                    currLoc = currLoc + spacing
            
            # Draw western tabs
            numTabs = self.sides[W].strength
            if numTabs > 0:
                spacing = h / (numTabs+1)
                currLoc = spacing
                for n in range(numTabs):
                    dc.DrawRectangle(0, int(currLoc-(tabSize/2)), tabSize, tabSize)
                    if not pdfCanvas is None:
                        pdfCanvas.rect(0, currLoc-(tabSize/2), tabSize, tabSize, stroke=1, fill=1)
                    elif not svgCanvas is None:
                        svgGroup.add(svgCanvas.rect(insert=(trans[0], trans[1]+currLoc-(tabSize/2)),
                         size=(tabSize, tabSize), fill='black', stroke='black', stroke_width=0))
                    currLoc = currLoc + spacing

            # Draw eastern tabs
            numTabs = self.sides[E].strength
            if numTabs > 0:
                spacing = h / (numTabs+1)
                currLoc = spacing
                for n in range(numTabs):
                    dc.DrawRectangle(w-tabSize, int(currLoc-(tabSize/2)), tabSize, tabSize)
                    if not pdfCanvas is None:
                        pdfCanvas.rect(w-tabSize, currLoc-(tabSize/2), tabSize, tabSize, stroke=1, fill=1)
                    elif not svgCanvas is None:
                        svgGroup.add(svgCanvas.rect(insert=(trans[0]+w-tabSize, trans[1]+currLoc-(tabSize/2)),
                         size=(tabSize, tabSize), fill='black', stroke='black', stroke_width=0))
                    currLoc = currLoc + spacing

            dc.SetTextForeground(self.textColorVector)
            if not pdfCanvas is None:
                pdfColor = (float(self.textColorVector[0])/255, float(self.textColorVector[1])/255, float(self.textColorVector[2])/255)
            if not svgCanvas is None:
                svgColor = "rgb(" + str(self.textColorVector[0]) + "," + str(self.textColorVector[1]) + "," + str(self.textColorVector[2]) + ")"

            # Draw tile label
            wLabel = (w - 2*tabSize)/2
            hLabel = (h - 2*tabSize)/2
            defPixHt = 30
            glueLabelScale = 1.0
            ftype = wx.FONTFAMILY_DEFAULT#wx.FONTFAMILY_MODERN
            fname = "Helvetica"#"Arial Narrow"
            pdfFont = "Helvetica"
            pdfFontBold = "Helvetica-Bold"
            
            #font = wx.Font(pointSize = 30, family = wx.DEFAULT, style = wx.NORMAL, weight = wx.NORMAL, faceName = 'Arial')
            if (not self.label is None) and ((mask == DISPLAY_ALL or mask == DISPLAY_LABEL)):
                font = wx.Font(pixelSize = (0,defPixHt), family = ftype, style = wx.NORMAL, weight = wx.FONTWEIGHT_BOLD, faceName = fname)#, faceName = 'Arial')
                font = GetFontSizeToFitTextInBounds(self.label, font, dc, wLabel, hLabel)
                pdfFontSize = font.GetPointSize()*1.2
                (wLabel, hLabel) = dc.GetTextExtent(self.label)
                dc.DrawText(self.label, int((w/2)-(wLabel/2)), int((h/2)-(hLabel/2)))
                
                if not pdfCanvas is None:
                    textobject = pdfCanvas.beginText((w/2)-(wLabel/2), (h/2)-(hLabel/2))
                    textobject.setFont(pdfFontBold, pdfFontSize)
                    textobject.setFillColorRGB(pdfColor[0], pdfColor[1], pdfColor[2])
                    textobject.textOut(self.label)
                    pdfCanvas.drawText(textobject)
                elif not svgCanvas is None:
                    text = svgCanvas.text(self.label,insert=(trans[0]+(w/2)-(wLabel/2), trans[1]+(h/2)+(hLabel/2)),
                                          font_size=font.GetPointSize()*svgPt,style="font-weight:bold;font-family:"+fname)#,textLength=textSize*units)
                    svgGroup.add(text)
            
            if (mask == DISPLAY_ALL) or (mask == DISPLAY_GLUES):
                longestLabel = ""
                for d in Dirs:
                    if len(self.sides[d].label) > len(longestLabel):
                        longestLabel = self.sides[d].label
    
                defWLabel = (w - 6*tabSize) * glueLabelScale
                defHLabel = (h - 6*tabSize) * glueLabelScale
                wLabel = defWLabel
                hLabel = defHLabel
                font = wx.Font(pointSize = 32, family = ftype, style = wx.NORMAL, weight = wx.FONTWEIGHT_NORMAL, faceName = fname)
                font = GetFontSizeToFitTextInBounds(longestLabel, font, dc, wLabel, hLabel)
                pdfFontSize = font.GetPointSize()
                
                if len(self.sides[N].label) > 0:
                    # Draw north glue label
                    wLabel = defWLabel
                    hLabel = defHLabel
                    (wLabel, hLabel) = dc.GetTextExtent(self.sides[N].label)
                    dc.DrawText(self.sides[N].label, int((w/2)-(wLabel/2)), 10)
                    
                    if not pdfCanvas is None:
                        textobject = pdfCanvas.beginText((w/2)-0.8*(wLabel/2), h-hLabel)
                        textobject.setFont(pdfFont, pdfFontSize)
                        textobject.setFillColorRGB(pdfColor[0], pdfColor[1], pdfColor[2])
                        textobject.textOut(self.sides[N].label)
                        pdfCanvas.drawText(textobject)
                    elif not svgCanvas is None:
                        text = svgCanvas.text(self.sides[N].label,insert=(trans[0]+(w/2)-(wLabel/2), trans[1]+1.3*hLabel),
                                              font_size=font.GetPointSize()*svgPt,style="font-family:"+fname)#,textLength=textSize*units)
                        svgGroup.add(text)
                
                if len(self.sides[S].label) > 0:
                    # Draw south glue label
                    wLabel = defWLabel
                    hLabel = defHLabel
                    (wLabel, hLabel) = dc.GetTextExtent(self.sides[S].label)
                    dc.DrawText(self.sides[S].label, int((w/2)-(wLabel/2)), h-hLabel)
                    
                    if not pdfCanvas is None:
                        textobject = pdfCanvas.beginText((w/2)-0.8*(wLabel/2), pdfFontSize)
                        textobject.setFont(pdfFont, pdfFontSize)
                        textobject.setFillColorRGB(pdfColor[0], pdfColor[1], pdfColor[2])
                        textobject.textOut(self.sides[S].label)
                        pdfCanvas.drawText(textobject)
                    elif not svgCanvas is None:
                        text = svgCanvas.text(self.sides[S].label,insert=(trans[0]+(w/2)-(wLabel/2), trans[1]+h-1.3*hLabel/2),
                                              font_size=font.GetPointSize()*svgPt,style="font-family:"+fname)#,textLength=textSize*units)
                        svgGroup.add(text)
                
                if len(self.sides[W].label) > 0:
                    # Draw west glue label
                    wLabel = defWLabel
                    hLabel = defHLabel
                    (wLabel, hLabel) = dc.GetTextExtent(self.sides[W].label)
                    dc.DrawRotatedText(self.sides[W].label, tabSize, int((h/2)+(wLabel/2)), 90.0)

                    if not pdfCanvas is None:
                        textobject = pdfCanvas.beginText((h/2)-0.8*(wLabel/2), -(tabSize+pdfFontSize))
                        textobject.setFont(pdfFont, pdfFontSize)
                        textobject.setFillColorRGB(pdfColor[0], pdfColor[1], pdfColor[2])
                        pdfCanvas.rotate(90)
                        textobject.textOut(self.sides[W].label)
                        pdfCanvas.drawText(textobject)
                        pdfCanvas.rotate(-90)
                    elif not svgCanvas is None:
                        text = svgCanvas.text(self.sides[W].label,insert=(-(trans[1]+(h/2)+(wLabel/2)), trans[0]+hLabel+tabSize),
                                              font_size=font.GetPointSize()*svgPt,style="font-family:"+fname)#,textLength=textSize*units)
                        text.rotate(270)
                        svgGroup.add(text)
    
                if len(self.sides[E].label) > 0:
                    # Draw east glue label
                    wLabel = defWLabel
                    hLabel = defHLabel
                    (wLabel, hLabel) = dc.GetTextExtent(self.sides[E].label)
                    dc.DrawRotatedText(self.sides[E].label, w-tabSize, int((h/2)-(wLabel/2)), 270.0)
                    
                    if not pdfCanvas is None:
                        textobject = pdfCanvas.beginText(-(h/2)-0.8*(wLabel/2), w-(tabSize+pdfFontSize))
                        textobject.setFont(pdfFont, pdfFontSize)
                        textobject.setFillColorRGB(pdfColor[0], pdfColor[1], pdfColor[2])
                        pdfCanvas.rotate(-90)
                        textobject.textOut(self.sides[E].label)
                        pdfCanvas.drawText(textobject)
                        pdfCanvas.rotate(90)
                    elif not svgCanvas is None:
                        text = svgCanvas.text(self.sides[E].label,insert=((trans[1]+(h/2)-(wLabel/2)), -(trans[0]+w-tabSize-hLabel)),
                                              font_size=font.GetPointSize()*svgPt,style="font-family:"+fname)#,textLength=textSize*units)
                        text.rotate(90)
                        svgGroup.add(text)

            dc.SelectObject(wx.NullBitmap)
            #img = bmp.ConvertToImage()
            #img = img.Scale(CELLSIZE, CELLSIZE, wx.IMAGE_QUALITY_HIGH)
            #bmp = wx.Bitmap(img)
            self.images[mask] = bmp

    def Copy(self):
        if self.dims == 2:
            t = PyTASTileType()
        else:
            t = TileType(True)
        t.sides[N] = self.sides[N].Copy()
        t.sides[E] = self.sides[E].Copy()
        t.sides[S] = self.sides[S].Copy()
        t.sides[W] = self.sides[W].Copy()
        if self.dims == 3:
            t.sides[U] = self.sides[U].Copy()
            t.sides[D] = self.sides[D].Copy()
        t.name = self.name
        t.label = self.label
        t.color = self.color
        t.textColor = self.textColor
        t.colorVector = self.colorVector
        t.textColorVector = self.textColorVector
        for imgNum in range(len(self.images)):
            if self.images[imgNum] is None:
                t.images[imgNum] = None
            else:
                t.images[imgNum] = self.images[imgNum].GetSubBitmap(wx.Rect(0, 0, self.images[imgNum].GetWidth(), self.images[imgNum].GetHeight()))
        return t
    
    def SetEqual(self, other):
        TileType.SetEqual(self, other)
        for imgNum in range(len(self.images)):
            if other.images[imgNum] is None:
                self.images[imgNum] = None
            else:
                self.images[imgNum] = other.images[imgNum].GetSubBitmap(wx.Rect(0, 0, other.images[imgNum].GetWidth(), other.images[imgNum].GetHeight()))

    def SetInputSides(self, inputSet):
        for k in inputSet:
            self.sides[k] = Glue(inputSet[k].strength, inputSet[k].label)

#----------------------------------------------------------------------------
class NeighborGlues():
    """Description of glues which border a frontier location, derived from neighbor tiles"""
    name = ""
    sides = {}
    dims  = 2
    
    def __init__(self, name="", is3D=False):
        self.name = name
        if is3D == False:
            self.dims = 2
            self.sides = {N:Glue(),E:Glue(),S:Glue(),W:Glue()}
        else:
            self.dims = 3
            self.sides = {N:Glue(),E:Glue(),S:Glue(),W:Glue(),U:Glue(),D:Glue()}
            
    def Print(self):
        print("Printing info for NeighborGlues:" + self.name + ":")
        printList = [N,E,S,W];
        if self.dims == 3:
            printList += [U,D]
        for face in printList:
            print("Face:" + str(face) + ":")
            if self.sides[face] is not None:
                print("Glue label:" + self.sides[face].label + ":")
            else:
                print("Glue is None")
    